import { TestingModule, Test } from '@nestjs/testing';
import { PrismaService } from '@modules/authentication/infrastructure/prisma.service';
import { EventEmitterModule } from '@nestjs/event-emitter';
import {
  DatabaseErrorException,
  NotFoundException,
  UniqueConstraintException,
} from '@mobicoop/ddd-library';
import {
  CreateUsernameProps,
  Type,
} from '@modules/authentication/core/domain/username.types';
import { UsernameRepository } from '@modules/authentication/infrastructure/username.repository';
import { UsernameMapper } from '@modules/authentication/username.mapper';
import * as bcrypt from 'bcrypt';
import { UsernameEntity } from '@modules/authentication/core/domain/username.entity';
import { AUTH_MESSAGE_PUBLISHER } from '@modules/authentication/authentication.di-tokens';

const authUuid = 'a4524d22-7be3-46cd-8444-3145470476dc';

const createUsernameProps: CreateUsernameProps = {
  userId: authUuid,
  type: Type.EMAIL,
  name: 'john.doe@email.com',
};

const mockMessagePublisher = {
  publish: jest.fn().mockImplementation(),
};

const mockLogger = {
  log: jest.fn(),
  warn: jest.fn(),
  error: jest.fn(),
};

describe('UsernameRepository', () => {
  let prismaService: PrismaService;
  let usernameRepository: UsernameRepository;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [EventEmitterModule.forRoot()],
      providers: [
        UsernameRepository,
        PrismaService,
        UsernameMapper,
        {
          provide: AUTH_MESSAGE_PUBLISHER,
          useValue: mockMessagePublisher,
        },
      ],
    })
      // disable logging
      .setLogger(mockLogger)
      .compile();

    prismaService = module.get<PrismaService>(PrismaService);
    usernameRepository = module.get<UsernameRepository>(UsernameRepository);
  });

  afterAll(async () => {
    await prismaService.$disconnect();
  });

  beforeEach(async () => {
    await prismaService.username.deleteMany();
    await prismaService.auth.deleteMany();
    await prismaService.auth.create({
      data: {
        uuid: authUuid,
        password: bcrypt.hashSync(`password`, 10),
      },
    });
  });

  describe('findByName', () => {
    it('should return a Username', async () => {
      await prismaService.username.create({
        data: {
          authUuid,
          username: 'john.doe@email.com',
          type: Type.EMAIL,
        },
      });

      const username = await usernameRepository.findByName(
        'john.doe@email.com',
      );
      expect(username.id).toBe('john.doe@email.com');
    });

    it('should throw an exception if record is not found', async () => {
      await expect(
        usernameRepository.findOne({
          username: 'jane.doe@email.com',
        }),
      ).rejects.toBeInstanceOf(NotFoundException);
    });
  });

  describe('create', () => {
    it('should create a username', async () => {
      const beforeCount = await prismaService.username.count();

      const usernameToCreate: UsernameEntity = await UsernameEntity.create(
        createUsernameProps,
      );
      await usernameRepository.insert(usernameToCreate);

      const afterCount = await prismaService.username.count();

      expect(afterCount - beforeCount).toBe(1);
    });

    it('should throw a UniqueConstraintException if username already exists', async () => {
      await prismaService.username.create({
        data: {
          authUuid,
          type: Type.EMAIL,
          username: 'john.doe@email.com',
        },
      });

      const usernameToCreate: UsernameEntity = await UsernameEntity.create(
        createUsernameProps,
      );
      await expect(
        usernameRepository.insert(usernameToCreate),
      ).rejects.toBeInstanceOf(UniqueConstraintException);
    });
  });

  describe('update username', () => {
    it('should update the name of a username', async () => {
      const usernameToUpdate = await prismaService.username.create({
        data: {
          authUuid,
          type: Type.EMAIL,
          username: 'johnny.doe@email.com',
        },
      });

      const toUpdate: UsernameEntity = await UsernameEntity.create(
        createUsernameProps,
      );
      await usernameRepository.updateUsername(
        usernameToUpdate.username,
        toUpdate,
      );

      const updatedUsername = await prismaService.username.findUnique({
        where: {
          username: toUpdate.id,
        },
      });

      expect(updatedUsername.username).toBe('john.doe@email.com');
      expect(usernameToUpdate.updatedAt.getTime()).toBeLessThan(
        updatedUsername.updatedAt.getTime(),
      );
    });

    it('should throw a DatabaseException if id is unknown', async () => {
      const toUpdate: UsernameEntity = await UsernameEntity.create(
        createUsernameProps,
      );

      await expect(
        usernameRepository.updateUsername('jane.doe@email.com', toUpdate),
      ).rejects.toBeInstanceOf(DatabaseErrorException);
    });
  });

  describe('delete', () => {
    it('should delete a username', async () => {
      await prismaService.username.create({
        data: {
          authUuid,
          type: Type.EMAIL,
          username: 'john.doe@email.com',
        },
      });
      const toDelete: UsernameEntity = await UsernameEntity.create(
        createUsernameProps,
      );
      await usernameRepository.deleteUsername(toDelete);

      const count = await prismaService.username.count();
      expect(count).toBe(0);
    });

    it('should throw a DatabaseException if username does not exist', async () => {
      const toDelete: UsernameEntity = await UsernameEntity.create(
        createUsernameProps,
      );
      await expect(usernameRepository.delete(toDelete)).rejects.toBeInstanceOf(
        DatabaseErrorException,
      );
    });
  });
});
