import { AggregateID, NotFoundException } from '@mobicoop/ddd-library';
import {
  AUTHENTICATION_REPOSITORY,
  USERNAME_REPOSITORY,
} from '@modules/authentication/authentication.di-tokens';
import { ValidateAuthenticationQuery } from '@modules/authentication/core/application/queries/validate-authentication/validate-authentication.query';
import { ValidateAuthenticationQueryHandler } from '@modules/authentication/core/application/queries/validate-authentication/validate-authentication.query-handler';
import { UnauthorizedException } from '@nestjs/common';

import { Test, TestingModule } from '@nestjs/testing';

const validName = 'john.doe@email.com';
const furtherInvalidName = 'validNameButInvalidAuthenticationForSomeReason';
const validUserId = '165192d4-398a-4469-a16b-98c02cc6f531';
const invalidUserId = 'b3b75798-7104-4151-adab-11fe16e32d85';
const validPassword = 'V@l1dP@$$w0rd';
const invalidPassword = '1nV@l1dP@$$w0rd';

const mockUsernameRepository = {
  findByName: jest.fn().mockImplementation((name: string) => {
    if (name == validName) {
      return {
        getProps: jest.fn().mockImplementation(() => ({
          userId: validUserId,
        })),
      };
    }
    if (name == furtherInvalidName) {
      return {
        getProps: jest.fn().mockImplementation(() => ({
          userId: invalidUserId,
        })),
      };
    }
    throw new Error();
  }),
};

const mockAuthenticationEntity = {
  id: validUserId,
  authenticate: jest
    .fn()
    .mockImplementation((password: string) => password === validPassword),
};

const mockAuthenticationRepository = {
  findOneById: jest.fn().mockImplementation((id: string) => {
    if (id === validUserId) return mockAuthenticationEntity;
    throw new Error();
  }),
};

describe('Validate Authentication Query Handler', () => {
  let validateAuthenticationQueryHandler: ValidateAuthenticationQueryHandler;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: AUTHENTICATION_REPOSITORY,
          useValue: mockAuthenticationRepository,
        },
        {
          provide: USERNAME_REPOSITORY,
          useValue: mockUsernameRepository,
        },
        ValidateAuthenticationQueryHandler,
      ],
    }).compile();

    validateAuthenticationQueryHandler =
      module.get<ValidateAuthenticationQueryHandler>(
        ValidateAuthenticationQueryHandler,
      );
  });

  it('should be defined', () => {
    expect(validateAuthenticationQueryHandler).toBeDefined();
  });

  describe('execution', () => {
    it('should validate an authentication query', async () => {
      const validateAuthenticationQuery = new ValidateAuthenticationQuery(
        validName,
        validPassword,
      );
      const id: AggregateID = await validateAuthenticationQueryHandler.execute(
        validateAuthenticationQuery,
      );
      expect(id).toBe(validUserId);
    });
    it('should not validate an authentication query if username is invalid', async () => {
      const validateAuthenticationQuery = new ValidateAuthenticationQuery(
        'invalidName',
        validPassword,
      );
      await expect(
        validateAuthenticationQueryHandler.execute(validateAuthenticationQuery),
      ).rejects.toBeInstanceOf(NotFoundException);
    });
    it('should not validate an authentication query if authentication is not found', async () => {
      const validateAuthenticationQuery = new ValidateAuthenticationQuery(
        furtherInvalidName,
        validPassword,
      );
      await expect(
        validateAuthenticationQueryHandler.execute(validateAuthenticationQuery),
      ).rejects.toBeInstanceOf(NotFoundException);
    });
    it('should not validate a failing authentication query', async () => {
      const validateAuthenticationQuery = new ValidateAuthenticationQuery(
        validName,
        invalidPassword,
      );
      await expect(
        validateAuthenticationQueryHandler.execute(validateAuthenticationQuery),
      ).rejects.toBeInstanceOf(UnauthorizedException);
    });
  });
});
