import { IdResponse } from '@mobicoop/ddd-library';
import { RpcExceptionCode } from '@mobicoop/ddd-library';
import { AuthenticationAlreadyExistsException } from '@modules/authentication/core/domain/authentication.errors';
import { Type } from '@modules/authentication/core/domain/username.types';
import { CreateAuthenticationGrpcController } from '@modules/authentication/interface/grpc-controllers/create-authentication.grpc.controller';
import { CreateAuthenticationRequestDto } from '@modules/authentication/interface/grpc-controllers/dtos/create-authentication.request.dto';
import { CommandBus } from '@nestjs/cqrs';
import { RpcException } from '@nestjs/microservices';
import { Test, TestingModule } from '@nestjs/testing';

const createAuthenticationRequest: CreateAuthenticationRequestDto = {
  userId: '78153e03-4861-4f58-a705-88526efee53b',
  password: 'John123',
  usernames: [
    {
      name: 'john.doe@email.com',
      type: Type.EMAIL,
    },
  ],
};

const mockCommandBus = {
  execute: jest
    .fn()
    .mockImplementationOnce(() => '78153e03-4861-4f58-a705-88526efee53b')
    .mockImplementationOnce(() => {
      throw new AuthenticationAlreadyExistsException();
    })
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Create Authentication Grpc Controller', () => {
  let createAuthenticationGrpcController: CreateAuthenticationGrpcController;

  beforeAll(async () => {
    const module: TestingModule = await Test.createTestingModule({
      providers: [
        {
          provide: CommandBus,
          useValue: mockCommandBus,
        },
        CreateAuthenticationGrpcController,
      ],
    }).compile();

    createAuthenticationGrpcController =
      module.get<CreateAuthenticationGrpcController>(
        CreateAuthenticationGrpcController,
      );
  });

  afterEach(async () => {
    jest.clearAllMocks();
  });

  it('should be defined', () => {
    expect(createAuthenticationGrpcController).toBeDefined();
  });

  it('should create a new authentication', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    const result: IdResponse = await createAuthenticationGrpcController.create(
      createAuthenticationRequest,
    );
    expect(result).toBeInstanceOf(IdResponse);
    expect(result.id).toBe('78153e03-4861-4f58-a705-88526efee53b');
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a dedicated RpcException if authentication already exists', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await createAuthenticationGrpcController.create(
        createAuthenticationRequest,
      );
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.ALREADY_EXISTS);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });

  it('should throw a generic RpcException', async () => {
    jest.spyOn(mockCommandBus, 'execute');
    expect.assertions(3);
    try {
      await createAuthenticationGrpcController.create(
        createAuthenticationRequest,
      );
    } catch (e: any) {
      expect(e).toBeInstanceOf(RpcException);
      expect(e.error.code).toBe(RpcExceptionCode.UNKNOWN);
    }
    expect(mockCommandBus.execute).toHaveBeenCalledTimes(1);
  });
});
